import sys
from os import mkdir
from os.path import exists, dirname, basename, join, split
from textx.metamodel import metamodel_from_file
from textx.export import metamodel_export
from textx.model import get_children_of_type
from textx.exceptions import TextXSemanticError
from textx.scoping import providers
from copy import copy, deepcopy

this_folder = dirname(__file__)

# set of actions used in the model
actions = set([])
# action alphabet of each state machine
alphabet = {}
# set of actions, per class, requiring synchronisation
syncactions = {}

# dictionary providing the channels to which a given port is connected
connected_channel = {}

# dictionary of array names in model (with types and sizes)
arraynames = {}

# set of statemachine names in the system
smnames = set([])
# map smnames to smobjects
smname_to_object = {}

def RepresentsInt(s):
    try: 
        int(s)
        return True
    except ValueError:
        return False

# Classes for SLCO statements
class Assignment(object):
	def __init__(self, parent, left, right):
		self.parent = parent
		self.left = left
		self.right = right

class Composite(object):
	def __init__(self, parent, guard, assignments):
		self.parent = parent
		self.guard = guard
		self.assignments = assignments

class Expression(object):
	def __init__(self, parent, left, op, right):
		self.parent = parent
		self.left = left
		self.op = op
		self.right = right

class ExprPrec1(object):
	def __init__(self, parent, left, op, right):
		self.parent = parent
		self.left = left
		self.op = op
		self.right = right

class ExprPrec2(object):
	def __init__(self, parent, left, op, right):
		self.parent = parent
		self.left = left
		self.op = op
		self.right = right

class ExprPrec3(object):
	def __init__(self, parent, left, op, right):
		self.parent = parent
		self.left = left
		self.op = op
		self.right = right

class ExprPrec4(object):
	def __init__(self, parent, left, op, right):
		self.parent = parent
		self.left = left
		self.op = op
		self.right = right

class Primary(object):
	def __init__(self, parent, sign, value, body, ref):
		self.parent = parent
		self.sign = sign
		self.value = value
		self.body = body
		self.ref = ref

class ExpressionRef(object):
	def __init__(self, parent, ref, index):
		self.parent = parent
		self.ref = ref
		self.index = index

class Variable(object):
	def __init__(self, parent, type, name, defvalue, defvalues):
		self.parent = parent
		self.type = type
		self.name = name
		self.defvalue = defvalue
		self.defvalues = defvalues

class VariableRef(object):
	def __init__(self, parent, var, index):
		self.parent = parent
		self.var = var
		self.index = index

class Type(object):
	def __init__(self, parent, base, size):
		self.parent = parent
		self.base = base
		self.size = size

class Action(object):
	def __init__(self, parent, name):
		self.parent = parent
		self.name = name

# extra class to identify references to Actions
class ActionRef(object):
	def __init__(self, parent, act):
		self.parent = parent
		self.act = act

# extra classes for AL-SLCO models
class AL_Assignment(object):
	def __init__(self, parent, left, right):
		self.parent = parent
		self.left = left
		self.right = right
		self.cached = cached

class AL_Expression(object):
	def __init__(self, parent, left, op, right):
		self.parent = parent
		self.left = left
		self.op = op
		self.right = right
		self.cached = cached

class AL_ReadInstruction(object):
	def __init__(self, parent, ref):
		self.parent = parent
		self.ref = ref

class WriteInstruction(object):
	def __init__(self, parent, ref):
		self.parent = parent
		self.ref = ref

# method to raise semantic error
def raise_semantic_error(S, s, model):
	"""S is error message string.
	s is object in which error occurs."""
	line, col = model._tx_parser.pos_to_linecol(s._tx_position)
	S += ' at ("%s", "%s")' % (line, col)
	raise TextXSemanticError(S)

# FUNCTIONS TO CREATE INSTANCES OF SLCO CONCEPTS

def create_smlocal_var(sm, v, type, size):
	"""Create and return a variable object for the variable named v of given type and size, associated to the given state machine sm"""
	newv = Variable(sm, '', v, None, [])
	if type == "Boolean":
		t = Type(newv, 'Boolean', size)
	elif type == "Byte":
		t = Type(newv, 'Byte', size)
	else:
		t = Type(newv, 'Boolean', size)
	newv.type = t
	# add variable to list of state machine-local variables
	sm.variables.append(newv)
	return newv

def create_var_expression(st, v, i):
	"""Create and return a new expression containing only a reference to the given variable name (with possibly index i). Associate the expression with the given statement st"""
	e = Expression(st, '', '', '')
	e4 = ExprPrec4(e, '', '', '')
	e3 = ExprPrec3(e4, '', '', '')
	e2 = ExprPrec2(e3, '', '', '')
	e1 = ExprPrec1(e2, '', '', '')
	p = Primary(e1, '', '', '', '')
	r = ExpressionRef(p, v, i)
	e.left = e4
	e4.left = e3
	e3.left = e2
	e2.left = e1
	e1.left = p
	p.ref = r
	return e

# FUNCTIONS TO EXTRACT INFORMATION FROM STATEMENTS

def scopename(v,i,o):
	"""Return full name of variable v, possibly with index i, if constant. in case a Port is given, return name of connected Channel"""
	if v.__class__.__name__ == "Channel":
		if i == None:
			name = v.name
		elif v.synctype == 'async':
			if RepresentsInt(i):
				name = v.name + "[" + str(i) + "][0]"
			else:
				name = v.name + str(i)
		else:
			name = i
	else:
		name = o.name
		if v.parent.__class__.__name__ == "StateMachine":
			name += "'" + v.parent.name
		name += "'" + v.name
		if v.__class__.__name__ != "StateMachine":
			if i != None:
				i_str = getlabel(i)
				if RepresentsInt(i_str):
					name += "[" + i_str + "]"
	return name

def sm_variables(sm):
	"""Provide list of variables accessible from current state machine sm"""
	Ldict = {}
	for v in sm.parent.variables:
		Ldict[v.name] = v
	for v in sm.variables:
		Ldict[v.name] = v
	return Ldict.values()

def statement_write_varrefs(s, o):
	"""Return a list of variable refs to which the statement is writing. o is Object owning s."""
	global connected_channel

	W = set([])
	if s.__class__.__name__ == "Assignment":
		W.add((s.left.var, s.left.index))
	elif s.__class__.__name__ == "Composite":
		for i in range(0,len(s.assignments)):
			W.add((s.assignments[i].left.var, s.assignments[i].left.index))
	elif s.__class__.__name__ == "ReceiveSignal":
		for i in range(0,len(s.params)):
			W.add((s.params[i].var, s.params[i].index))
	return W

def statement_read_varrefs(s, o, sm, only_unguarded):
	"""Return a set of variable refs from which the statement is reading. o is Object owning s. sm is the state machine owning s. only_unguarded indicates whether only the unguarded objects (true) or all objects (false) should be returned."""
	global connected_channel, smname_to_object, smnames, alphabet
	R = set([])
	if s.__class__.__name__ == "Assignment":
		if s.left.index != None:
			R |= statement_read_varrefs(s.left.index, o, sm, only_unguarded)
		R |= statement_read_varrefs(s.right, o, sm, only_unguarded)
	elif s.__class__.__name__ == "Composite":
		if s.guard != None:
			R |= statement_read_varrefs(s.guard, o, sm, only_unguarded)
		if s.guard == None or not only_unguarded:
			for i in range(0,len(s.assignments)):
				R |= statement_read_varrefs(s.assignments[i], o, sm, only_unguarded)
	elif s.__class__.__name__ == "SendSignal":
		if not only_unguarded:
			for p in s.params:
				R |= statement_read_varrefs(p, o, sm, only_unguarded)
	elif s.__class__.__name__ == "ReceiveSignal":
		c = connected_channel[(o, s.target)]
		if c.synctype == 'async':
			if not only_unguarded:
				# add other elements of buffer head element
				for i in range(1,len(c.type)+1):
				 	R.add((c, i))
		if not only_unguarded:
			Messagerefs = []
			R1 = set([])
			for p in s.params:
				Messagerefs.append((p.var.name, getlabel(p.index)))
				if p.index != None:
					R1 |= statement_read_varrefs(p.index, o, sm, only_unguarded)
			R1 |= statement_read_varrefs(s.guard, o, sm, only_unguarded)
			for (v,j) in R1:
				j_str = getlabel(j)
				found = False
				for (v2name,j2_str) in Messagerefs:
					if v.name == v2name and j_str == j2_str:
						found = True
						break
				if not found:
					R.add((v,j))
	elif s.__class__.__name__ == "Expression" or s.__class__.__name__ == "ExprPrec4" or s.__class__.__name__ == "ExprPrec3" or s.__class__.__name__ == "ExprPrec2" or s.__class__.__name__ == "ExprPrec1":
		R |= statement_read_varrefs(s.left, o, sm, only_unguarded)
		if s.op != '':
			R |= statement_read_varrefs(s.right, o, sm, only_unguarded)
	elif s.__class__.__name__ == "Primary":
		if s.ref != None:
			if s.ref.ref in actions:
				# we have a user-defined action
				if s.ref.ref in syncactions.get(o.type,set([])):
					# add references to other state machines in Object o that have the action in their alphabet
					# (i.e., we add a reference to their current state variable)
					for sname in smnames:
						(o2,sm2) = smname_to_object[sname]
						if o == o2 and sm != sm2:
							if s.ref.ref in alphabet[sm2]:
								R.add((sm2,"src"))
			if s.ref.index != None:
				R |= statement_read_varrefs(s.ref.index, o, sm, only_unguarded)
			# obtain suitable object matching name s.ref.ref
			for v1 in sm_variables(sm):
				if v1.name == s.ref.ref:
					R.add((v1, s.ref.index))
					break
		if s.body != None:
			R |= statement_read_varrefs(s.body, o, sm, only_unguarded)
	elif s.__class__.__name__ == "VariableRef":
		if s.index != None:
			R |= statement_read_varrefs(s.index, o, sm, only_unguarded)
		R.add((s.var, s.index))
	return R

def statement_has_array_read_access(s, o, sm, only_unguarded):
	"""Return whether statement s reads from at least one array."""
	global arraynames
	R = statement_read_varrefs(s, o, sm, only_unguarded)
	for (v,i) in R:
		result = arraynames.get(scopename(v, None, o), -1)
		if result != -1:
			return True
	return False

def statement_has_array_access(s, o, sm):
	"""Return whether statement s reads from or writes to at least one array."""
	global arraynames
	R = statement_read_varrefs(s, o, sm, False)
	R |= statement_write_varrefs(s, o)
	for (v,i) in R:
		result = arraynames.get(scopename(v, None, o), -1)
		if result != -1:
			return True
	return False

# *** MODEL PROCESSORS ***

# model processor to create the set of actions, and alphabet per statemachine dictionary
def construct_action_sets(model, metamodel):
	global actions, alphabet, syncactions
	# construct the set of actions as they appear in the model
	# actions = set([])
	# for a in model.actions:
	# 	actions.add(a.name)
            
    # construct dictionary providing target states when performing a given action in a given state
	Adict = {}
	actiontargets = {}
	actions = set([])
	for c in model.classes:
		for sm in c.statemachines:
			A = set([])
			smtrans = {}
			for t in sm.transitions:
				for st in t.statements:
					if statement_is_actionref(st):
						a = getlabel(st)
						A.add(a)
						atrans = smtrans.get(a, {})
						src = state_id[t.source]
						tgt = state_id[t.target]
						tgts = atrans.get(src, [])
						tgts.append(tgt)
						tgts = list(set(tgts))
						atrans[src] = tgts
						smtrans[a] = atrans
			actiontargets[sm] = smtrans
			for a in A:
				actions.add(a)
				C = Adict.get(a, {})
				count = C.get(c,0)
				count += 1
				C[c] = count
				Adict[a] = C
                
	# construct set of actions requiring synchronisation
	syncactions = {}
	for a in actions:
		C = Adict[a]
		for c in C.keys():
			if C[c] > 1:
				CC = syncactions.get(c,set([]))
				CC.add(a)
				syncactions[c] = CC

# model processor to construct dictionary providing connected channel of given port
def construct_connected_channel_dict(model, metamodel):
	global connected_channel
	connected_channel = {}
	for c in model.channels:
		connected_channel[(c.source, c.port0)] = c
		connected_channel[(c.target, c.port1)] = c

# model processor to create dictionary of array names
def construct_arraynames_dict(model, metamodel):
	global arraynames
	arraynames = {}
	for o in model.objects:
		for v in o.type.variables:
			if v.type.size > 0:
				arraynames[o.name + "'" + v.name] = (v.type, v.type.size)
		for sm in o.type.statemachines:
			for v in sm.variables:
				if v.type.size > 0:
					arraynames[o.name + "'" + sm.name + "'" + v.name] = (v.type, v.type.size)

# model processor to construct set of statemachine names in the system, and to construct a map from names to objects
def construct_smnames_to_objects_dict(model, metamodel):
	global smnames, smname_to_object
	smnames = set([])
	smname_to_object = {}
	for o in model.objects:
		for sm in o.type.statemachines:
			smnames.add(o.name + "'" + sm.name)
			smname_to_object[o.name + "'" + sm.name] = (o,sm)
	#print("here" + str(smnames))

def getlabel_colored(s):
	"""Get the label for the given statement s"""
	result = ''
	if s.__class__.__name__ == "Assignment":
		result += s.left.var.name
		if s.left.index != None:
			result += "[" + getlabel(s.left.index) + "]"
		result += " := " + getlabel(s.right)
	elif s.__class__.__name__ == "Composite":
		result += "[ "
		if s.guard != None:
			result += getlabel(s.guard)
			result += "; "
		for i in range(0,len(s.assignments)):
			result += getlabel(s.assignments[i])
			if i < len(s.assignments)-1:
				result += "; "
		result += " ]"
	elif s.__class__.__name__ == "Delay":
		result += "<b>after </b>" + str(s.length) + "<b> ms</b>"
	elif s.__class__.__name__ == "SendSignal":
		result += "<b>send </b>" + s.signal + "("
		first = True
		for p in s.params:
			if not first:
				result += ","
			else:
				first = False
			result += getlabel(p)
		result += ") <b>to </b>" + s.target.name
	elif s.__class__.__name__ == "ReceiveSignal":
		result += "<b>receive </b>" + s.signal + "("
		first = True
		for p in s.params:
			if not first:
				result += ","
			else:
				first = False
			result += getlabel(p)
		if s.guard != None:
			result += " | " + getlabel(s.guard)
		result += ") <b>from </b>" + s.target.name
	elif s.__class__.__name__ == "Expression" or s.__class__.__name__ == "ExprPrec4" or s.__class__.__name__ == "ExprPrec3" or s.__class__.__name__ == "ExprPrec2" or s.__class__.__name__ == "ExprPrec1":
		if s.op != '':
			result += getlabel(s.left) + " " + s.op + " " + getlabel(s.right)
		else:
			result += getlabel(s.left)
	elif s.__class__.__name__ == "Primary":
		result += s.sign
		if s.sign == "not":
			result += " "
		if s.value != None:
			newvalue = s.value
			result += str(newvalue)
		elif s.ref != None:
			result += s.ref.ref
			if s.ref.index != None:
				result += "[" + getlabel(s.ref.index) + "]"
		else:
			result += '(' + getlabel(s.body) + ')'
	elif s.__class__.__name__ == "VariableRef":
		result += s.var.name
		if s.index != None:
			result += "[" + getlabel(s.index) + "]"
	return result

def getlabel(s):
	"""Get the label for the given statement s"""
	result = ''
	if s.__class__.__name__ == "Assignment":
		result += s.left.var.name
		if s.left.index != None:
			result += "[" + getlabel(s.left.index) + "]"
		result += " := " + getlabel(s.right)
	elif s.__class__.__name__ == "Composite":
		result += "[ "
		if s.guard != None:
			result += getlabel(s.guard)
			result += "; "
		for i in range(0,len(s.assignments)):
			result += getlabel(s.assignments[i])
			if i < len(s.assignments)-1:
				result += "; "
		result += " ]"
	elif s.__class__.__name__ == "Delay":
		result += "after " + str(s.length) + " ms"
	elif s.__class__.__name__ == "SendSignal":
		result += "send " + s.signal + "("
		first = True
		for p in s.params:
			if not first:
				result += ","
			else:
				first = False
			result += getlabel(p)
		result += ") to " + s.target.name
	elif s.__class__.__name__ == "ReceiveSignal":
		result += "receive " + s.signal + "("
		first = True
		for p in s.params:
			if not first:
				result += ","
			else:
				first = False
			result += getlabel(p)
		if s.guard != None:
			result += " | " + getlabel(s.guard)
		result += ") from " + s.target.name
	elif s.__class__.__name__ == "Expression" or s.__class__.__name__ == "ExprPrec4" or s.__class__.__name__ == "ExprPrec3" or s.__class__.__name__ == "ExprPrec2" or s.__class__.__name__ == "ExprPrec1":
		if s.op != '':
			result += getlabel(s.left) + " " + s.op + " " + getlabel(s.right)
		else:
			result += getlabel(s.left)
	elif s.__class__.__name__ == "Primary":
		result += s.sign
		if s.sign == "not":
			result += " "
		if s.value != None:
			newvalue = s.value
			result += str(newvalue)
		elif s.ref != None:
			result += s.ref.ref
			if s.ref.index != None:
				result += "[" + getlabel(s.ref.index) + "]"
		else:
			result += '(' + getlabel(s.body) + ')'
	elif s.__class__.__name__ == "VariableRef":
		result += s.var.name
		if s.index != None:
			result += "[" + getlabel(s.index) + "]"
	elif s.__class__.__name__ == "Variable":
		result += s.name
	elif s.__class__.__name__ == "ActionRef":
		result += s.act.name
	return result

# model processor to check for name clashes and remove duplicates in lists
# invalid name clashes:
# - action names cannot be used for variable names
# - classes need to have unique names
# - objects need to have unique names
def check_names(model, metamodel):
	# actions
	actlist = []
	actset = set([])
	for a in model.actions:
		if a.name not in actset:
			actset.add(a.name)
			actlist.append(a)
	model.actions = actlist

	# classes
	cnames = set([])
	for c in model.classes:
		if c.name in cnames:
			error = 'Name clash: "%s" used for multiple classes, once' % name
			raise_semantic_error(error, c, model)
		else:
			cnames.add(c.name)

	# check inside each class for variable names
	for c in model.classes:
		# variables
		varlist = []
		varset = set([])
		for v in c.variables:
			if v.name in actset:
				error = 'Name clash: "%s" used for both a variable and an action ' % v.name
				raise_semantic_error(error, v, model)
			else:
				if v.name not in varset:
					varset.add(v.name)
					varlist.append(v)
		c.variables = varlist
		# ports
		portlist = []
		tmp = set([])
		for p in c.ports:
			if p.name not in tmp:
				tmp.add(p.name)
				portlist.append(p)
		c.ports = portlist
		# state machines
		for sm in c.statemachines:
			varlist = []
			varset = set([])
			for v in sm.variables:
				if v.name in actset:
					error = 'Name clash: "%s" used for both a variable and an action at' % v.name
					raise_semantic_error(error, v, model)
				else:
					if v.name not in varset:
						varset.add(v.name)
						varlist.append(v)
			sm.variables = varlist
			# states
			statelist = []
			stateset = set([sm.initialstate.name])
			for s in sm.states:
				if s.name not in stateset:
					stateset.add(s.name)
					statelist.append(s)
			sm.states = statelist

	onames = set([])
	for o in model.objects:
		if o.name in cnames:
			error = 'Name clash: "%s" used for multiple objects, once' % o.name
			raise_semantic_error(error, o, model)
		else:
			onames.add(o.name)

# model processor adding initial state to list of states
def add_initial_to_states(model, metamodel):
	for c in model.classes:
		for stm in c.statemachines:
			stm.states = [stm.initialstate] + stm.states

# model processor filling in types of variables
def add_variable_types(model, metamodel):
	for c in model.classes:
		for i in range(0,len(c.variables)):
			if c.variables[i].type == None:
				c.variables[i].type = c.variables[i-1].type
		for sm in c.statemachines:
			for i in range(0,len(sm.variables)):
				if sm.variables[i].type == None:
					sm.variables[i].type = sm.variables[i-1].type

# model processor setting sizes of types by default to one
def set_default_type_size(model, metamodel):
	for c in model.classes:
		for i in range(0,len(c.variables)):
			if c.variables[i].type != None:
				if c.variables[i].type.size != None:
					if c.variables[i].type.size == 0:
						c.variables[i].type.size = 1
		for sm in c.statemachines:
			for i in range(0,len(sm.variables)):
				if sm.variables[i].type != None:
					if sm.variables[i].type.size != None:
						if sm.variables[i].type.size == 0:
							sm.variables[i].type.size = 1
	for ch in model.channels:
		for t in ch.type:
			if t.size != None:
				if t.size == 0:
					t.size = 1

# model processor setting sizes of channels by default to one
def set_default_channel_size(model, metamodel):
	for ch in model.channels:
		if ch.size == 0:
			ch.size = 1

# model processor adding tau action to transitions without statements
def add_taus(model, metamodel):
	tau_needed = False
	for c in model.classes:
		for stm in c.statemachines:
			for trn in stm.transitions:
				if len(trn.statements) == 0:
					tau_needed = True
					break
			if tau_needed:
				break
		if tau_needed:
			break
	if tau_needed:
		# add tau to list of actions
		ta = Action(model, "tau")
		model.actions.append(ta)
		# add reference to tau action to all transitions without statements
		for c in model.classes:
			for stm in c.statemachines:
				for trn in stm.transitions:
					if len(trn.statements) == 0:
						trn.statements.append(ActionRef(trn, ta))

# model processor to fix and check references
def fix_references(model, metamodel):
	for c in model.classes:
		Vc = {}
		for v in c.variables:
			Vc[v.name] = v
		for sm in c.statemachines:
			V = copy(Vc)
			for v in sm.variables:
				V[v.name] = v
			statedict = {sm.initialstate.name: sm.initialstate}
			for s in sm.states:
				statedict[s.name] = s
			for tr in sm.transitions:
				# check state references of transitions
				sref = statedict.get(tr.source.name)
				if sref == None:
					error = 'Source state "%s" mentioned in transition is not defined in state machine' % tr.source.name
					raise_semantic_error(error, tr, model)
				else:
					tr.source = sref
				sref = statedict.get(tr.target.name)
				if sref == None:
					error = 'Target state "%s" mentioned in transition is not defined in state machine' % tr.target.name
					raise_semantic_error(error, tr, model)
				else:
					tr.target = sref
				# check references to actions and variables in statements
				for st in tr.statements:
					statement_check_refs(st, V, model)

# # model processor to check initialisation of variables: TODO
# def check_inits(model, metamodel):
# 	for o in model.objects:
# 		for i in o.assignments:
# 			t = i.left.type
# 			if t.base == 'Integer' and t.size == 0 and not i.right == 

def statement_check_refs(s, V, model):
	"""Auxiliary function used to check references in statements.
	V is a dictionary of references to variables in the current scope."""
	global actions

	if s.__class__.__name__ == "Assignment":
		statement_check_refs(s.left, V, model)
		statement_check_refs(s.right, V, model)
	elif s.__class__.__name__ == "Composite":
		if s.guard != None:
			statement_check_refs(s.guard, V, model)
		for a in s.assignments:
			statement_check_refs(a, V, model)
	elif s.__class__.__name__ == "ReceiveSignal":
		for p in s.params:
			statement_check_refs(p, V, model)
		if s.guard != None:
			statement_check_refs(s.guard, V, model)
	elif s.__class__.__name__ == "SendSignal":
		for p in s.params:
			statement_check_refs(p, V, model)
	elif s.__class__.__name__ == "VariableRef":
		vref = V.get(s.var.name)
		if vref == None:
			error = 'There is a reference to a variable "%s" that does not exist in that scope' % s.var.name
			raise_semantic_error(error, s, model)
		else:
			# possibly fix reference to take scope into account
			s.var = vref
			# if variable is of Array type, an index must be provided
			if vref.type.size > 1 and s.index == None:
				error = 'There is a reference to an Array variable "%s" where an index is mandatory, but missing,' % s.var.name
				raise_semantic_error(error, s, model)
	elif s.__class__.__name__ != "Primary" and s.__class__.__name__ != "ActionRef":
		statement_check_refs(s.left, V, model)
		if s.op != '':
			statement_check_refs(s.right, V, model)
	elif s.__class__.__name__ != "ActionRef":
		if s.ref != None:
			if s.ref.ref not in actions:
				vref = V.get(s.ref.ref)
				if vref == None:
					error = 'There is a reference to a variable "%s" that does not exist in that scope' % s.ref.ref
					raise_semantic_error(error, s, model)
				else:
					# if variable is of Array type, an index must be provided
					if vref.type.size > 1 and s.ref.index == None:
						error = 'There is a reference to an Array variable "%s" where an index is mandatory, but missing,' % s.var.name
						raise_semantic_error(error, s, model)

# model processor to simplify action references
#def simplify_statements(model, metamodel):
#	for c in model.classes:
#		for sm in c.statemachines:
#			for tr in sm.transitions:
#				for st in tr.statements:
#					if expression_is_actionref(st):
						# replace with a single class of type ActionRef
						# TODO

def statement_is_actionref(s):
	if s.__class__.__name__ == "ActionRef":
		return True
	else:
		return expression_is_actionref(s)

def expression_is_actionref(s):
	"""Determine whether the given expression is an action reference"""
	global actions
	if s.__class__.__name__ == "Expression":
		if s.op == '':
			snext = s.left
			if snext.op == '':
				snext = snext.left
				if snext.op == '':
					snext = snext.left
					if snext.op == '':
						snext = snext.left
						if snext.op == '':
							snext = snext.left
							if snext.ref != None and snext.sign == '':
								snext = snext.ref
								if snext.ref in actions:
									return True
	return False

def read_SLCO_model(m):
	"""Read, post process, and type check an SLCO model"""

	# create meta-model
	slco_mm = metamodel_from_file(join(this_folder,'../textx_grammars/slco2.tx'), autokwd=True, classes=[Assignment, Composite, Expression, ExprPrec1, ExprPrec2, ExprPrec3, ExprPrec4, Primary, ExpressionRef, Variable, VariableRef, Type, Action])

	# register processors
	slco_mm.register_model_processor(construct_action_sets)
	slco_mm.register_model_processor(construct_connected_channel_dict)
	slco_mm.register_model_processor(construct_arraynames_dict)
	slco_mm.register_model_processor(construct_smnames_to_objects_dict)
	slco_mm.register_model_processor(check_names)
	slco_mm.register_model_processor(add_initial_to_states)
	slco_mm.register_model_processor(add_variable_types)
	#slco_mm.register_model_processor(set_default_type_size)
	slco_mm.register_model_processor(set_default_channel_size)
	slco_mm.register_model_processor(add_taus)
	slco_mm.register_model_processor(fix_references)
	#slco_mm.register_model_processor(simplify_statements)

# To do: Check receive statements for not receiving multiple values in the same variable
# To do: Check for absence of arrays (not single elements) as part of messages

	slco_mm.register_scope_providers({
		"*.*": providers.FQN(),
		"Initialisation.left": providers.RelativeName("parent.type.variables"),
		"Channel.port0": providers.RelativeName("source.type.ports"),
		"Channel.port1": providers.RelativeName("target.type.ports"),
		"ReceiveSignal.from": providers.RelativeName("parent.parent.parent.type.ports"),
		"SendSignal.to": providers.RelativeName("parent.parent.parent.type.ports"),
	})

	# parse and return the model
	return slco_mm.model_from_file(m)
